@@ -14,6 +14,11 @@ import sys
 import traceback
 import unittest
 
+try:
+    import argcomplete
+except ImportError:
+    argcomplete = None
+
 from tests import virtinstall, virtclone, virtconvert, virtxml
 from tests import utils
 
@@ -21,9 +26,10 @@ os.environ["LANG"] = "en_US.UTF-8"
 os.environ["HOME"] = "/tmp"
 os.environ["DISPLAY"] = ":3.4"
 
-OLD_OSINFO = utils.has_old_osinfo()
 TMP_IMAGE_DIR = "/tmp/__virtinst_cli_"
 XMLDIR = "tests/cli-test-xml"
+OLD_OSINFO = utils.has_old_osinfo()
+HAS_ISOINFO = find_executable("isoinfo")
 
 # Images that will be created by virt-install/virt-clone, and removed before
 # each run
@@ -44,6 +50,7 @@ exist_images = [
 iso_links = [
     "/tmp/fake-fedora17-tree.iso",
     "/tmp/fake-centos65-label.iso",
+    "/tmp/fake-no-osinfo.iso",
 ]
 
 exist_files = exist_images
@@ -70,11 +77,22 @@ test_files = {
     'EXISTIMG4': exist_images[1],
     'ISOTREE': iso_links[0],
     'ISOLABEL': iso_links[1],
+    'ISO-NO-OS': iso_links[2],
     'TREEDIR': "%s/fakefedoratree" % XMLDIR,
     'COLLIDE': "/dev/default-pool/collidevol1.img",
 }
 
 
+def has_old_osinfo():
+    if OLD_OSINFO:
+        return "osinfo is too old"
+
+
+def missing_isoinfo():
+    if not HAS_ISOINFO:
+        return "isoinfo not installed"
+
+
 ######################
 # Test class helpers #
 ######################
@@ -89,24 +107,32 @@ class Command(object):
         self.compare_file = None
         self.input_file = None
 
-        self.skip_check = None
+        self.need_conn = True
+        self.skip_cb = None
         self.check_version = None
         self.grep = None
+        self.nogrep = None
 
         app, opts = self.cmdstr.split(" ", 1)
         self.app = app
         self.argv = [os.path.abspath(app)] + shlex.split(opts)
+        self.env = None
 
     def _launch_command(self, conn):
         logging.debug(self.cmdstr)
 
         app = self.argv[0]
 
+        oldenv = None
         oldstdout = sys.stdout
         oldstderr = sys.stderr
         oldstdin = sys.stdin
         oldargv = sys.argv
         try:
+            if self.env:
+                oldenv = os.environ.copy()
+                os.environ.update(self.env)
+
             out = io.StringIO()
 
             sys.stdout = out
@@ -142,6 +168,8 @@ class Command(object):
             sys.stderr = oldstderr
             sys.stdin = oldstdin
             sys.argv = oldargv
+            if oldenv:
+                os.environ = oldenv
 
 
     def _get_output(self, conn):
@@ -181,12 +209,12 @@ class Command(object):
                     conn = utils.URIs.openconn(self.argv[idx + 1])
                     break
 
-            if not conn:
+            if not conn and self.need_conn:
                 raise RuntimeError("couldn't parse URI from command %s" %
                                    self.argv)
 
-            if self.skip_check:
-                tests.skipTest("skipped with skip_check")
+            if self.skip_cb and self.skip_cb():
+                tests.skipTest("skip_cb: %s" % self.skip_cb())
                 return
 
             code, output = self._get_output(conn)
@@ -205,6 +233,9 @@ class Command(object):
 
             if self.grep and self.grep not in output:
                 _raise_error("Didn't find grep=%s" % self.grep)
+            if self.nogrep and self.nogrep in output:
+                _raise_error("Found grep=%s when we shouldn't see it" %
+                        self.nogrep)
 
             if self.compare_file:
                 if self._check_support(tests, conn, self.check_version,
@@ -238,12 +269,12 @@ class Command(object):
 
 
 class _CategoryProxy(object):
-    def __init__(self, app, name, default_args, skip_check, check_version):
+    def __init__(self, app, name, default_args, skip_cb, check_version):
         self._app = app
         self._name = name
 
         self.default_args = default_args
-        self.skip_check = skip_check
+        self.skip_cb = skip_cb
         self.check_version = check_version
 
     def add_valid(self, *args, **kwargs):
@@ -292,14 +323,14 @@ class App(object):
 
 
     def add_category(self, catname, default_args,
-                     skip_check=None, check_version=None):
+                     skip_cb=None, check_version=None):
         obj = _CategoryProxy(self, catname, default_args,
-                             skip_check, check_version)
+                             skip_cb, check_version)
         self.categories[catname] = obj
         return obj
 
     def _add(self, catname, testargs, valid, compfile,
-             skip_check=None, check_version=None, input_file=None,
+             skip_cb=None, check_version=None, input_file=None,
              auto_printarg=True, grep=None):
 
         category = self.categories[catname]
@@ -314,7 +345,7 @@ class App(object):
             compfile = os.path.basename(self.appname) + "-" + compfile
             compare_XMLDIR = "%s/compare" % XMLDIR
             cmd.compare_file = "%s/%s.xml" % (compare_XMLDIR, compfile)
-        cmd.skip_check = skip_check or category.skip_check
+        cmd.skip_cb = skip_cb or category.skip_cb
         cmd.check_version = (check_version or
                              category.check_version or
                              self.check_version)
@@ -351,7 +382,7 @@ vinst = App("virt-install")
 # virt-install verbose XML comparison tests #
 #############################################
 
-c = vinst.add_category("xml-comparsion", "--connect %(URI-KVM)s --noautoconsole --os-variant fedora-unknown", skip_check=OLD_OSINFO)
+c = vinst.add_category("xml-comparsion", "--connect %(URI-KVM)s --noautoconsole --os-variant fedora-unknown", skip_cb=has_old_osinfo)
 
 # Singleton element test #1, for simpler strings
 c.add_compare(""" \
@@ -401,7 +432,7 @@ cache.mode=emulate,cache.level=3 \
 --metadata title=my-title,description=my-description,uuid=00000000-1111-2222-3333-444444444444 \
 --boot cdrom,fd,hd,network,menu=off,loader=/foo/bar,emulator=/new/emu,bootloader=/new/bootld,rebootTimeout=3 \
 --idmap uid_start=0,uid_target=1000,uid_count=10,gid_start=0,gid_target=1000,gid_count=10 \
---security type=static,label='system_u:object_r:svirt_image_t:s0:c100,c200',relabel=yes \
+--security type=static,label='system_u:object_r:svirt_image_t:s0:c100,c200',relabel=yes,baselabel=baselabel \
 --numatune 1-3,4,mode=strict \
 --memtune hard_limit=10,soft_limit=20,swap_hard_limit=30,min_guarantee=40 \
 --blkiotune weight=100,device_path=/home/test/1.img,device_weight=200 \
@@ -527,6 +558,8 @@ c.add_compare(""" \
 \
 --panic iobase=507 \
 \
+--vsock cid=17 \
+\
 --qemu-commandline env=DISPLAY=:0.1 \
 --qemu-commandline="-display gtk,gl=on" \
 --qemu-commandline="-device vfio-pci,addr=05.0,sysfsdev=/sys/class/mdev_bus/0000:00:02.0/f321853c-c584-4a6b-b99a-3eee22a3919c" \
@@ -648,7 +681,7 @@ c.add_valid("--hvm --cdrom %(EXISTIMG1)s")  # Simple cdrom install
 c.add_valid("--wait 0 --os-variant winxp --cdrom %(EXISTIMG1)s")  # Windows (2 stage) install
 c.add_valid("--pxe --virt-type test")  # Explicit virt-type
 c.add_valid("--arch i686 --pxe")  # Explicitly fullvirt + arch
-c.add_valid("--location %(TREEDIR)s")  # Directory tree URL install
+c.add_valid("--location location=%(TREEDIR)s")  # Directory tree URL install
 c.add_valid("--location %(TREEDIR)s --initrd-inject virt-install --extra-args ks=file:/virt-install")  # initrd-inject
 c.add_valid("--hvm --location %(TREEDIR)s --extra-args console=ttyS0")  # Directory tree URL install with extra-args
 c.add_valid("--paravirt --location %(TREEDIR)s")  # Paravirt location
@@ -667,6 +700,7 @@ c.add_invalid("--pxe --os-variant farrrrrrrge")  # Bogus --os-variant
 c.add_invalid("--pxe --boot menu=foobar")
 c.add_invalid("--cdrom %(EXISTIMG1)s --extra-args console=ttyS0")  # cdrom fail w/ extra-args
 c.add_invalid("--hvm --boot kernel=%(TREEDIR)s/pxeboot/vmlinuz,initrd=%(TREEDIR)s/pxeboot/initrd.img,kernel_args='foo bar' --initrd-inject virt-install")  # initrd-inject with manual kernel/initrd
+c.add_invalid("--disk none --location kernel=/dev/null,initrd=/dev/null")  # --location with manual kernel/initrd, but not URL
 
 c = vinst.add_category("single-disk-install", "--nographics --noautoconsole --disk %(EXISTIMG1)s")
 c.add_valid("--hvm --import")  # FV Import install
@@ -706,12 +740,12 @@ c.add_invalid("--file /foo/bar/baz --pxe")  # Trying to use unmanaged storage wi
 ###########################
 
 c = vinst.add_category("kvm-generic", "--connect %(URI-KVM)s --noautoconsole")
-c.add_compare("--os-variant fedora-unknown --file %(EXISTIMG1)s --location %(TREEDIR)s --extra-args console=ttyS0 --cpu host --channel none --console none --sound none --redirdev none", "kvm-fedoralatest-url", skip_check=OLD_OSINFO)  # Fedora Directory tree URL install with extra-args
+c.add_compare("--os-variant fedora-unknown --file %(EXISTIMG1)s --location %(TREEDIR)s --extra-args console=ttyS0 --cpu host --channel none --console none --sound none --redirdev none", "kvm-fedoralatest-url", skip_cb=has_old_osinfo)  # Fedora Directory tree URL install with extra-args
 c.add_compare("--test-media-detection %(TREEDIR)s", "test-url-detection")  # --test-media-detection
-c.add_compare("--os-variant fedora20 --disk %(NEWIMG1)s,size=.01,format=vmdk --location %(TREEDIR)s --extra-args console=ttyS0 --quiet", "quiet-url", skip_check=OLD_OSINFO)  # Quiet URL install should make no noise
+c.add_compare("--os-variant fedora20 --disk %(NEWIMG1)s,size=.01,format=vmdk --location %(TREEDIR)s --extra-args console=ttyS0 --quiet", "quiet-url", skip_cb=has_old_osinfo)  # Quiet URL install should make no noise
 c.add_compare("--cdrom %(EXISTIMG2)s --file %(EXISTIMG1)s --os-variant win2k3 --wait 0 --sound --controller usb", "kvm-win2k3-cdrom")  # HVM windows install with disk
 c.add_compare("--os-variant ubuntusaucy --nodisks --boot cdrom --virt-type qemu --cpu Penryn --input tablet", "qemu-plain")  # plain qemu
-c.add_compare("--os-variant fedora20 --nodisks --boot network --nographics --arch i686", "qemu-32-on-64", skip_check=OLD_OSINFO)  # 32 on 64
+c.add_compare("--os-variant fedora20 --nodisks --boot network --nographics --arch i686", "qemu-32-on-64", skip_cb=has_old_osinfo)  # 32 on 64
 
 # ppc64 tests
 c.add_compare("--arch ppc64 --machine pseries --boot network --disk %(EXISTIMG1)s --disk device=cdrom --os-variant fedora20 --network none", "ppc64-pseries-f20")
@@ -719,29 +753,30 @@ c.add_compare("--arch ppc64 --boot network --disk %(EXISTIMG1)s --os-variant fed
 c.add_compare("--connect %(URI-KVM-PPC64LE)s --import --disk %(EXISTIMG1)s --os-variant fedora20 --panic default", "ppc64le-kvm-import")
 
 # s390x tests
-c.add_compare("--arch s390x --machine s390-ccw-virtio --connect %(URI-KVM-S390X)s --boot kernel=/kernel.img,initrd=/initrd.img --disk %(EXISTIMG1)s --disk %(EXISTIMG3)s,device=cdrom --os-variant fedora21", "s390x-cdrom", skip_check=OLD_OSINFO)
-c.add_compare("--arch s390x --machine s390-ccw-virtio --connect " + utils.URIs.kvm_s390x_KVMIBM + " --boot kernel=/kernel.img,initrd=/initrd.img --disk %(EXISTIMG1)s --disk %(EXISTIMG3)s,device=cdrom --os-variant fedora21 --watchdog diag288,action=reset --panic default", "s390x-cdrom-KVMIBM")
+c.add_compare("--arch s390x --machine s390-ccw-virtio --connect %(URI-KVM-S390X)s --boot kernel=/kernel.img,initrd=/initrd.img --disk %(EXISTIMG1)s --disk %(EXISTIMG3)s,device=cdrom --os-variant fedora21", "s390x-cdrom", skip_cb=has_old_osinfo)
+c.add_compare("--arch s390x --machine s390-ccw-virtio --connect " + utils.URIs.kvm_s390x_KVMIBM + " --boot kernel=/kernel.img,initrd=/initrd.img --disk %(EXISTIMG1)s --disk %(EXISTIMG3)s,device=cdrom --os-variant fedora21 --watchdog diag288,action=reset --panic default --graphics vnc", "s390x-cdrom-KVMIBM")
 
 # qemu:///session tests
-c.add_compare("--connect " + utils.URIs.kvm_session + " --disk size=8 --os-variant fedora21 --cdrom %(EXISTIMG1)s", "kvm-session-defaults", skip_check=OLD_OSINFO)
+c.add_compare("--connect " + utils.URIs.kvm_session + " --disk size=8 --os-variant fedora21 --cdrom %(EXISTIMG1)s", "kvm-session-defaults", skip_cb=has_old_osinfo)
 
 # misc KVM config tests
-c.add_compare("--disk %(EXISTIMG1)s --location %(ISOTREE)s --nonetworks", "location-iso", skip_check=not find_executable("isoinfo"))  # Using --location iso mounting
+c.add_compare("--disk none --location %(ISO-NO-OS)s,kernel=frib.img,initrd=/frob.img", "location-manual-kernel", skip_cb=missing_isoinfo)  # --location with an unknown ISO but manually specified kernel paths
+c.add_compare("--disk %(EXISTIMG1)s --location %(ISOTREE)s --nonetworks", "location-iso", skip_cb=missing_isoinfo)  # Using --location iso mounting
 c.add_compare("--disk %(EXISTIMG1)s --cdrom %(ISOLABEL)s", "cdrom-centos-label")  # Using --cdrom with centos CD label, should use virtio etc.
 c.add_compare("--disk %(EXISTIMG1)s --pxe --os-variant rhel5.4", "kvm-rhel5")  # RHEL5 defaults
 c.add_compare("--disk %(EXISTIMG1)s --pxe --os-variant rhel6.4", "kvm-rhel6")  # RHEL6 defaults
-c.add_compare("--disk %(EXISTIMG1)s --pxe --os-variant rhel7.0", "kvm-rhel7", skip_check=OLD_OSINFO)  # RHEL7 defaults
-c.add_compare("--connect " + utils.URIs.kvm_nodomcaps + " --disk %(EXISTIMG1)s --pxe --os-variant rhel7.0", "kvm-cpu-default-fallback", skip_check=OLD_OSINFO)  # No domcaps, so mode=host-model isn't safe, so we fallback to host-model-only
+c.add_compare("--disk %(EXISTIMG1)s --pxe --os-variant rhel7.0", "kvm-rhel7", skip_cb=has_old_osinfo)  # RHEL7 defaults
+c.add_compare("--connect " + utils.URIs.kvm_nodomcaps + " --disk %(EXISTIMG1)s --pxe --os-variant rhel7.0", "kvm-cpu-default-fallback", skip_cb=has_old_osinfo)  # No domcaps, so mode=host-model isn't safe, so we fallback to host-model-only
 c.add_compare("--connect " + utils.URIs.kvm_nodomcaps + " --cpu host-copy --disk none --pxe", "kvm-hostcopy-fallback")  # No domcaps so need to use capabilities for CPU host-copy
-c.add_compare("--disk %(EXISTIMG1)s --pxe --os-variant centos7.0", "kvm-centos7", skip_check=OLD_OSINFO)  # Centos 7 defaults
-c.add_compare("--disk %(EXISTIMG1)s --pxe --os-variant centos7.0", "kvm-centos7", skip_check=OLD_OSINFO)  # Centos 7 defaults
-c.add_compare("--disk %(EXISTIMG1)s --cdrom %(EXISTIMG2)s --os-variant win10", "kvm-win10", skip_check=OLD_OSINFO)  # win10 defaults
-c.add_compare("--os-variant win7 --cdrom %(EXISTIMG2)s --boot loader_type=pflash,loader=CODE.fd,nvram_template=VARS.fd --disk %(EXISTIMG1)s", "win7-uefi", skip_check=OLD_OSINFO)  # no HYPER-V with UEFI
+c.add_compare("--disk %(EXISTIMG1)s --pxe --os-variant centos7.0", "kvm-centos7", skip_cb=has_old_osinfo)  # Centos 7 defaults
+c.add_compare("--disk %(EXISTIMG1)s --pxe --os-variant centos7.0", "kvm-centos7", skip_cb=has_old_osinfo)  # Centos 7 defaults
+c.add_compare("--disk %(EXISTIMG1)s --cdrom %(EXISTIMG2)s --os-variant win10", "kvm-win10", skip_cb=has_old_osinfo)  # win10 defaults
+c.add_compare("--os-variant win7 --cdrom %(EXISTIMG2)s --boot loader_type=pflash,loader=CODE.fd,nvram_template=VARS.fd --disk %(EXISTIMG1)s", "win7-uefi", skip_cb=has_old_osinfo)  # no HYPER-V with UEFI
 c.add_compare("--arch i686 --boot uefi --pxe --disk none", "kvm-i686-uefi")  # i686 uefi
 c.add_compare("--machine q35 --cdrom %(EXISTIMG2)s --disk %(EXISTIMG1)s", "q35-defaults")  # proper q35 disk defaults
 c.add_compare("--disk size=20 --os-variant solaris10", "solaris10-defaults")  # test solaris OS defaults, triggers a couple specific code paths
 c.add_compare("--disk size=1 --os-variant openbsd4.9", "openbsd-defaults")  # triggers net fallback scenario
-c.add_compare("--connect " + utils.URIs.kvm_remote + " --import --disk %(EXISTIMG1)s --os-variant fedora21 --pm suspend_to_disk=yes", "f21-kvm-remote", skip_check=OLD_OSINFO)
+c.add_compare("--connect " + utils.URIs.kvm_remote + " --import --disk %(EXISTIMG1)s --os-variant fedora21 --pm suspend_to_disk=yes", "f21-kvm-remote", skip_cb=has_old_osinfo)
 
 c.add_valid("--arch aarch64 --nodisks --pxe --connect " + utils.URIs.kvm_nodomcaps)  # attempt to default to aarch64 UEFI, but it fails, but should only print warnings
 c.add_invalid("--disk none --boot network --machine foobar")  # Unknown machine type
@@ -776,6 +811,7 @@ c.add_compare("--connect %(URI-KVM-AARCH64)s --disk none --network none --os-var
 
 c = vinst.add_category("lxc", "--name foolxc --memory 64 --noautoconsole --connect " + utils.URIs.lxc)
 c.add_compare("", "default")
+c.add_compare("--os-variant fedora27", "default-f27")
 c.add_compare("--filesystem /source,/", "fs-default")
 c.add_compare("--init /usr/bin/httpd", "manual-init")
 
@@ -848,7 +884,6 @@ c.add_invalid("--pxe --noautoconsole --wait 1", grep="Installation has exceeded
 c.add_valid("--pxe --nographics --transient", grep="testsuite console command: ['virsh'")  # --transient handling
 
 
-
 ##################
 # virt-xml tests #
 ##################
@@ -948,12 +983,18 @@ c.add_compare("--add-device --sound pcspk", "add-sound")
 c.add_compare("--add-device --disk %(EXISTIMG1)s,bus=virtio,target=vdf", "add-disk-basic")
 c.add_compare("--add-device --disk %(EXISTIMG1)s", "add-disk-notarget")  # filling in acceptable target
 c.add_compare("--add-device --disk %(NEWIMG1)s,size=.01", "add-disk-create-storage")
+c.add_compare("--add-device --disk size=.01", "add-disk-default-storage")
 c.add_compare("--remove-device --sound ich6", "remove-sound-model", check_version="1.3.5")  # check_version=new graphics listen output
 c.add_compare("--remove-device --disk 3", "remove-disk-index")
 c.add_compare("--remove-device --disk /dev/null", "remove-disk-path")
 c.add_compare("--remove-device --video all", "remove-video-all", check_version="1.3.3")  # check_version=video primary= attribute
 c.add_compare("--remove-device --host-device 0x04b3:0x4485", "remove-hostdev-name", check_version="1.2.11")  # check_version=video ram output change
 
+c = vixml.add_category("add/rm devices OS KVM", "--connect %(URI-KVM)s test --print-diff --define")
+c.add_compare("--add-device --disk %(EXISTIMG1)s", "kvm-add-disk-os-from-xml")  # Guest OS (none) from XML
+c.add_compare("--add-device --disk %(EXISTIMG1)s --os-variant fedora28", "kvm-add-disk-os-from-cmdline")  # Guest OS (fedora) provided on command line
+c.add_compare("--add-device --network default", "kvm-add-network-os-from-xml")  # Guest OS information taken from the guest XML
+c.add_compare("--add-device --network default --os-variant fedora28", "kvm-add-network-os-from-cmdline")  # Guest OS information provided on the command line
 
 
 
@@ -1026,6 +1067,44 @@ c.add_compare(_VMX_IMG + " --disk-format qcow2 --print-xml", "vmx-compare")
 c.add_compare(_OVF_IMG + " --disk-format none --destination /tmp --print-xml", "ovf-compare")
 
 
+#################################
+# argparse/autocomplete testing #
+#################################
+
+ARGCOMPLETE_CMDS = []
+
+
+def _add_argcomplete_cmd(line, grep, nogrep=None):
+    env = {
+        "_ARGCOMPLETE": "1",
+        "COMP_TYPE": "9",
+        "COMP_POINT": str(len(line)),
+        "COMP_LINE": line,
+        "_ARGCOMPLETE_COMP_WORDBREAKS": "\"'><;|&(:",
+    }
+    cmd = Command(line)
+    cmd.grep = grep
+    if nogrep:
+        cmd.nogrep = nogrep
+    cmd.env = env
+    cmd.need_conn = False
+    def have_argcomplete():
+        if not argcomplete:
+            return "argcomplete not installed"
+    cmd.skip_cb = have_argcomplete
+    ARGCOMPLETE_CMDS.append(cmd)
+
+_add_argcomplete_cmd("virt-install --di", "--disk")
+_add_argcomplete_cmd("virt-install --disk ", "driver.copy_on_read=")  # will list all --disk subprops
+_add_argcomplete_cmd("virt-install --disk a", "address.base")
+_add_argcomplete_cmd("virt-install --disk address.u", "address.unit")
+_add_argcomplete_cmd("virt-install --disk address.unit=foo,sg", "sgio")
+_add_argcomplete_cmd("virt-install --check d", "disk_size")
+_add_argcomplete_cmd("virt-install --test-stub", None,
+        nogrep="--test-stub-command")
+_add_argcomplete_cmd("virt-clone --preserve", "--preserve-data")
+_add_argcomplete_cmd("virt-xml --sound mode", "model")
+_add_argcomplete_cmd("virt-convert --dest", "--destination")
 
 
 #########################
@@ -1080,6 +1159,7 @@ _cmdlist += vinst.cmds
 _cmdlist += vclon.cmds
 _cmdlist += vconv.cmds
 _cmdlist += vixml.cmds
+_cmdlist += ARGCOMPLETE_CMDS
 
 # Generate numbered names like testCLI%d
 for _cmd in _cmdlist:
