@@ -2,7 +2,6 @@
 # Common code for all guests
 #
 # Copyright 2006-2009, 2013, 2014, 2015 Red Hat, Inc.
-# Jeremy Katz <katzj@redhat.com>
 #
 # This work is licensed under the GNU GPLv2 or later.
 # See the COPYING file in the top-level directory.
@@ -29,7 +28,7 @@ class _DomainDevices(XMLBuilder):
             'smartcard', 'serial', 'parallel', 'console', 'channel',
             'input', 'tpm', 'graphics', 'sound', 'video', 'hostdev',
             'redirdev', 'watchdog', 'memballoon', 'rng', 'panic',
-            'memory']
+            'memory', 'vsock']
 
 
     disk = XMLChildProperty(DeviceDisk)
@@ -53,6 +52,7 @@ class _DomainDevices(XMLBuilder):
     rng = XMLChildProperty(DeviceRng)
     panic = XMLChildProperty(DevicePanic)
     memory = XMLChildProperty(DeviceMemory)
+    vsock = XMLChildProperty(DeviceVsock)
 
     def get_all(self):
         retlist = []
@@ -475,11 +475,10 @@ class Guest(XMLBuilder):
     def set_defaults(self, _guest):
         if not self.uuid:
             self.uuid = util.generate_uuid(self.conn)
-        if not self.vcpus:
-            self.vcpus = 1
 
         self.set_capabilities_defaults()
 
+        self._set_default_resources()
         self._set_default_machine()
         self._set_default_uefi()
 
@@ -510,6 +509,16 @@ class Guest(XMLBuilder):
     # Private xml routines #
     ########################
 
+    def _set_default_resources(self):
+        res = self.osinfo.get_recommended_resources(self)
+
+        if not self.memory and res and res.get('ram') > 0:
+            self.memory = res['ram'] // 1024
+
+        if not self.vcpus:
+            self.vcpus = res.get('n-cpus') if res and res.get('n-cpus') > 0 else 1
+
+
     def _set_default_machine(self):
         if self.os.machine:
             return
